#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "ledcomm.h"
#include "util_macros.h"

#define NOP4() asm volatile("nop\n nop");

#ifndef __AVR_ATtiny2313__
#define NOPULLUPS() do { set_bit(SFIOR, PUD); } while(0)
#define PULLUPS() do { unset_bit(SFIOR, PUD); } while(0)
#else
#define NOPULLUPS() do { set_bit(MCUCR, PUD); } while(0)
#define PULLUPS() do { unset_bit(MCUCR, PUD); } while(0)
#endif

/* Idle loop: period for transmit & receive */
#define IDLE_TX 10  // 1ms
#define IDLE_RX 40  // 4ms

#define SYNC_TX    10  // 1ms
#define SYNC_RX    40  // 4ms
#define SYNC_DELAY 10  // 1ms

#define SPACE_TIME 5   // 0.5ms

// States
enum
{
    // System states
    STATE_Discovery,
    STATE_Sync,
    
    // Sub-states:    
    // LED is emitting or sensing
    LED_Send,
    LED_Receive,
    
    // Discovery period- maybe saw light and definitely saw light
    DISC_NotSeen,
    DISC_MaybeSeen,
    DISC_LightSeen,
    
    // STATE_Sync substates: RX, TX, waiting for TX
    SYNC_FindFallingEdge,
    SYNC_LedOn,
    SYNC_WaitBeforeTx,
    
    // Transfer: other unit is idling, or, we are receiving a frame from
    // other unit
    RECV_Idle,
    RECV_RecvFrame,

    XMIT_Idle,
    XMIT_StartBit,
    XMIT_Frame,
    XMIT_StopBit,

    RX_LightSeen,
    RX_LightNotSeen
} states;

/************************************/
/* Variables                        */
/************************************/
// Synchronized interrupt time
char INT_SYNC;

// State variables
volatile char System;    // State of the system
volatile char LED;       // State of the LED (sending / reading)
volatile char Disc;      // Discovery status
volatile char Sync;      // Syncing status
volatile char Recv;      // Transfer status (idle, sending or receiving)
volatile char Xmit;

// Timing variables
unsigned char sync_time; // Timekeeper for on/off LED pulses

// Counting variables
unsigned char light_count;   // Number of light measurements seen

// For falling edge detection
char falledge_last;  // Last bit received
unsigned char falledge_light; // Length of the most recent consecutive run of light-seen measurements
unsigned char falledge_dark;  // Length of the most recent consecutive run of light-not-seen measurements

// How many falling edges we've seen
volatile unsigned char sync_ctr;


volatile unsigned char rxbuf[LEDCOMM_MAXRXBUF];
volatile unsigned char rxbuf_pos;

volatile unsigned char rxbytebuf;
volatile unsigned char txbytebuf;
volatile unsigned char tx_nbits; // number of bits transmitted for current byte
volatile unsigned char rx_nbits; // number of bits received for current byte


/* Internal functions */

void LED_sense() {
    /* Charge first */

    /* Reverse bias: set TX pin output LOW, set RX pin output HIGH */
    set_bit(DDRLED, LEDRX);
    unset_bit(PORTLED, LEDTX);
    set_bit(PORTLED, LEDRX);
    NOP4();

    /* Discharge: set TX pin output LOW, set RX pin input, no pullup */
    NOPULLUPS();
    unset_bit(DDRLED, LEDRX);
    unset_bit(PORTLED, LEDRX);
    PULLUPS();
}

void LED_xmit() {
    /* Emit: set TX pin output HIGH, set RX pin output LOW */
    set_bit(PORTLED, LEDTX);
    //set_bit(DDRLED, LEDTX); /* Probably unnecessary --cgd */
    unset_bit(PORTLED, LEDRX);
    set_bit(DDRLED, LEDRX);

}

void LED_off() {
    unset_bit(PORTLED, LEDTX);
    // 
    //set_bit(DDRLED, LEDTX);
    unset_bit(PORTLED, LEDRX);
    set_bit(DDRLED, LEDRX);
}

/* For manipulating the RX and TX buffers */

/* Adds the single bit "bit" to the RX buffer */
void add_to_rx(unsigned char bit) {
    rxbytebuf |= (bit << rx_nbits);
    rx_nbits++;

    if(rx_nbits == 8) {
	if(rxbuf_pos < LEDCOMM_MAXRXBUF) {
	    rxbuf[rxbuf_pos] = rxbytebuf;
	}
	rxbytebuf = 0;
	rxbuf_pos++;
	rx_nbits = 0;

	Recv = RECV_Idle;
    }
}

/* Returns the current bit that needs to be TX'd */
unsigned char get_tx_bit() {
    return (txbytebuf & 0x01);
}

/* Returns -1 if you need to TX a stop bit (i.e. we're done with this byte) */
char next_tx_bit() {
    txbytebuf = txbytebuf >> 1;
    tx_nbits++;
    if(tx_nbits == 8) {
	// SAVESPACE: Shoudln't be necessary
	//tx_nbits = 0;
	return -1;
    }
    return 1;
}

/* The interrupt tick */
inline void ledcomm_tick() {
    if(System == STATE_Discovery) {
	// Discovery state-- LED on for 1ms, read for 4ms
	
	// LED is on
	if(LED == LED_Send) {
	    // Keep the LED on for 1ms
	    if(sync_time++ >= IDLE_TX) {
		// Afterwards, precharge the LED for receive
		LED = LED_Receive;
		sync_time = 0;
		
		// Set up the LED to read
		LED_sense();
	    }
	}
            
	// LED is receiving
	else {
	    // Read for 4ms
	    if(sync_time++ < IDLE_RX) {
		// Read the value of the pin before we charge up
		if(bit_is_clear(PINLED, LEDRX)) {
		    // We saw light this time around
		    // Saw light from an LED previously, time to sync
			
		    if(Disc == DISC_MaybeSeen) {
			// Update states
			System = STATE_Sync;
			Disc = DISC_LightSeen;
			Sync = SYNC_WaitBeforeTx;
			// SAVESPACE: Shouldn't be necessary
			//Recv = RECV_Idle;
			// Set the counters
			sync_time = 0;
			sync_ctr = 0;
                            
			return;
		    }
                        
		    // First time seeing light, come around again
		    Disc = DISC_MaybeSeen;
		} else {
		    // We didn't see light this time around
		    // Saw light before but not now-- possibly a fluke
		    if(Disc == DISC_MaybeSeen)
			Disc = DISC_NotSeen;
			
		}
                    
		// Set up the LED to read
		LED_sense();
	    }
                
	    // After the read cycle is done, go back to turning the LED on
	    else {
		// Configure the LED for output
		LED_xmit();
		    
		LED = LED_Send;
		sync_time = 0;
	    }
	}
    } else {
        
	// Synchronization state. Wait 1ms, then transmit a data bit.
	// Then look for a falling edge (and try to decode a data
	// bit).  Max read here will be 4ms, if no falling edge is
	// found in this amount of time, go back to discovery state
	    
	// LED is on
	if(Sync == SYNC_LedOn) {
	    // Keep the LED on for 1ms or 0.5ms
	    // Afterwards, charge the LED for receive
	    if(++sync_time >= SYNC_TX) {
		sync_time = 0;
		Sync = SYNC_FindFallingEdge;
		falledge_dark = 0;
		falledge_light = 0;
		falledge_last = RX_LightNotSeen;
		light_count = 0;

		/* We've sent a MARK; update transmitter state */
		if(Xmit == XMIT_StopBit) {
		    Xmit = XMIT_Idle;
		}
		    
		if(Xmit == XMIT_Frame) {
		    if (next_tx_bit() == -1) {
			Xmit = XMIT_StopBit;
		    }
		}

		if(Xmit == XMIT_StartBit) {
		    Xmit = XMIT_Frame;
		}
		    
		// Set up the LED to read
		LED_sense();
		    
	    } else if(sync_time >= SPACE_TIME) {
		/* So far we've sent a SPACE; do we have to stop? */
		if(Xmit == XMIT_StartBit) {
		    LED_off();
		} else if (Xmit == XMIT_Frame && get_tx_bit() == 0) {
		    LED_off();
		}
	    }
	    //unset_bit(PORTB, 4);
	    //unset_bit(PORTB, 5);
	} else if (Sync == SYNC_FindFallingEdge) {
	    if(++sync_time >= SYNC_RX) {
		// If we got here, we didn't see a falling edge in our RX window
		// We've lost contact
		// go to State_Discovery / DISC_NotSeen
		// TODO Discard a partially-received byte

		sync_time = 0;
		sync_ctr = 0;
		System = STATE_Discovery;
		Disc = DISC_NotSeen;

	    }
	    if(bit_is_clear(PINLED, LEDRX)) {
		// Light seen

		// Update falling-edge-counter state
		if(falledge_last == RX_LightSeen) {
		    falledge_light++;
		} else {
		    falledge_light = 1;
		}
		falledge_last = RX_LightSeen;

		light_count++;

	    } else {
		// Light not seen

		// Update falling-edge-counter state
		if(falledge_last == RX_LightNotSeen) {
		    falledge_dark++;
		} else {
		    falledge_dark = 1;
		}
		falledge_last = RX_LightNotSeen;


		// Check for a falling edge
		if(falledge_dark >= 8 && falledge_light >= 2) {
		    // We found a falling edge
		    // End measurement cycle; go to WaitBeforeTx
		    // TODO FIXME Does not delaying here hurt us?  Check this.
		    sync_time = 0;
		    Sync = SYNC_WaitBeforeTx;
		    if (sync_ctr < 5) sync_ctr++;
			
		    if (sync_ctr >= 3) {
			// Attempt to decode a data bit
			if (light_count <= 7) {
			    // got a SPACE
			    if(Recv == RECV_Idle) {
				// This SPACE marks the start of a frame
				rx_nbits = rxbytebuf = 0;
				Recv = RECV_RecvFrame;
			    } else {
				// or, this SPACE is in the middle of a frame
				//unset_bit(PORTB, 5);
				//set_bit(PORTB, 4);
				add_to_rx(0);
			    }
			} else {
			    // got a MARK
			    if(Recv == RECV_RecvFrame) {
				add_to_rx(1);
				//unset_bit(PORTB, 4);
				//set_bit(PORTB, 5);
			    }
			}
		    }

		    return;
		}
	    }

	    LED_sense();
	} else if (Sync == SYNC_WaitBeforeTx) {
	    if(++sync_time >= SYNC_DELAY) {
		// Go to TX
		sync_time = 0;
		Sync = SYNC_LedOn;
		LED_xmit();
	    }
	}
            

         
    }
}

/* API functions */
void ledcomm_init() {
    // Initialize the ports
    set_bit(DDRLED, LEDTX);
    set_bit(DDRLED, LEDRX);
    LED_off();

        
    // Initialize the timers and counters
    sync_time = 0;
    light_count = 0;
    
    // Initialize the states
    System = STATE_Discovery;   // Discovery mode
    LED = LED_Send;             // LED turned on 
    Disc = DISC_NotSeen;        // Haven't seen any light yet
    Sync = SYNC_WaitBeforeTx;         // Unsynchronized
    Recv = RECV_Idle;      // Other unit ain't sending anything
    Xmit = XMIT_Idle;
    
    rxbuf_pos = 0;
    rxbytebuf = txbytebuf = tx_nbits = rx_nbits = 0;

    // SAVESPACE: Shouldn't be necessary
    //memset((char *) rxbuf, 0, LEDCOMM_MAXRXBUF);

}

uint8_t ledcomm_client_present() {
    return sync_ctr >= 3;
}

unsigned char ledcomm_rx_len() {
    return rxbuf_pos;
}

volatile char* ledcomm_get_rxbuf() {
    return rxbuf;
}

void ledcomm_rx_reset() {
    rxbuf[0] = 0;
    rxbuf_pos = 0;
}

int8_t ledcomm_send_buffer(char* buf, uint8_t buflen) {
    uint8_t nbytes = 0;

    if(sync_ctr < 3) return -1;


    while(nbytes < buflen) {

	cli();
	tx_nbits = 0;
	txbytebuf = buf[nbytes];
	Xmit = XMIT_StartBit;
	sei();

	/* Wait to transmit the character */
	while(Xmit != XMIT_Idle && System == STATE_Sync) {}

	/* We lost contact while transmitting this byte, d'oh */
	if(System != STATE_Sync) return -1;

	nbytes++;
    }

    return 1;
}



void ledcomm_sleep() {
	LED_off();
}

void ledcomm_unsleep() {
}

